/*
 * ---------------- www.spacesimulator.net --------------
 *   ---- Space simulators and 3d engine tutorials ----
 *
 * Author: Damiano Vitulli
 *
 * This program is released under the BSD licence
 * By using this program you agree to licence terms on spacesimulator.net copyright page
 *
 *
 * Porting to cocoa by: Ariel Feinerman on 3/24/10 (arielfeinerman@gmail.com).
 *
 */

/* 
 * Spaceships credits:
 * fighter1.3ds - created by: Dario Vitulli 
 * fighter2.3ds - created by: Rene Reiter <renereiter@hotmail.com>
 * fighter3.3ds - taken from http://www.3dcafe.com/asp/meshes.asp
 */ 

#import <OpenGL/OpenGL.h>
#import <OpenGL/glu.h>

#import "AFController.h"
#import "AFTimer.h"
#import "Scene.h"

#import "object.h"
#import "camera.h"

/**********************************************************
 *
 * VARIABLES DECLARATION
 *
 *********************************************************/

// Absolute rotation values (0-359 degrees) and rotation increments for each frame
double rotation_x = 0, rotation_x_increment = 0.1;
double rotation_y = 0, rotation_y_increment = 0.05;
double rotation_z = 0, rotation_z_increment = 0.03;

// Flag for rendering as lines or filled polygons
int filling = 1; //0=OFF 1=ON

//Lights settings
GLfloat light_ambient[]= { 0.1f, 0.1f, 0.1f, 0.1f };
GLfloat light_diffuse[]= { 1.0f, 1.0f, 1.0f, 0.0f };
GLfloat light_specular[]= { 1.0f, 1.0f, 1.0f, 0.0f };
GLfloat light_position[]= { 100.0f, 0.0f, -10.0f, 1.0f };

//Materials settings
GLfloat mat_ambient[]= { 0.2f, 0.2f, 0.2f, 0.0f };
GLfloat mat_diffuse[]= { 0.2f, 0.2f, 0.2f, 0.0f };
GLfloat mat_specular[]= { 1.0f, 1.0f, 1.0f, 0.0f };
GLfloat mat_shininess[]= { 1.0f };

obj_type object[MAX_OBJECTS];
int obj_qty = 0; //Number of objects in our world 
int obj_control = 0; //Number of the object that we can control 

camera_type camera; //The camera 

@implementation Scene

- (id) initWithController: (AFController *) theController {
	
	if (theController == nil) {
		[self release];
		return nil;
	}
	
	if (self = [super init]) {
		controller = theController;
	}
	
	return self;
}

- (void) change {
	
	[self render];
}

- (void) keyDown: (NSEvent *) theEvent {
	
	unichar unicodeKey;
	unicodeKey = [[theEvent characters] characterAtIndex: 0];
	
	switch (unicodeKey) {
		
		case 'f':
		case 'F':
			if (![theEvent isARepeat] &&
				([theEvent modifierFlags] & NSCommandKeyMask)) {
				(FullScreen) ? ([controller exitFullScreen]) : ([controller enterFullScreen]);
			}
			break;
			
		case 'j': 
		case 'J':
			ObjTranslate(&object[obj_control], 0, 0, -1);
			break;
			
		case 'm': 
		case 'M':
			ObjTranslate(&object[obj_control], 0, 0, 1);
			break;
			
        case 'k': 
		case 'K':
			ObjRotate(&object[obj_control], 0, 20, 0);
			break;
			
        case 'l': 
		case 'L':
			ObjRotate(&object[obj_control], 0, -20, 0);
			break;
			
        case 'e': 
		case 'E':
			CamRotate(&camera, 20, 0, 0);
			break;
			
        case 'c': 
		case 'C':
			CamRotate(&camera, -20, 0, 0);
			break;
			
        case 'a': 
		case 'A':
			CamRotate(&camera, 0, -20, 0);
			break;
        
		case 'd': 
		case 'D':
			CamRotate(&camera, 0, 20, 0);
			break;
			
        case 'z': 
		case 'Z':
			CamRotate(&camera, 0, 0, -20);
			break;
			
        case 'x': 
		case 'X':
			CamRotate(&camera, 0, 0, 20);
			break;
			
        case 'w': 
		case 'W':
			CamTranslate(&camera, 0, 0, 1);
			break;
			
        case 's': 
		case 'S':
			CamTranslate(&camera, 0, 0, -1);
			break;
			
		case 'r': 
		case 'R':
            if (filling == 0)
            {
                glPolygonMode (GL_FRONT_AND_BACK, GL_FILL); // Polygon rasterization mode (polygon filled)
                filling = 1;
            }   
            else 
            {
                glPolygonMode (GL_FRONT_AND_BACK, GL_LINE); // Polygon rasterization mode (polygon outlined)
                filling = 0;
            }
			break;
			
		case NSUpArrowFunctionKey:
			ObjRotate(&object[obj_control], -20, 0, 0);
			break;

		case NSDownArrowFunctionKey:
			ObjRotate(&object[obj_control], 20, 0, 0);
			break;

		case NSLeftArrowFunctionKey:
			ObjRotate(&object[obj_control], 0, 0, 20);
			break;

		case NSRightArrowFunctionKey:
			ObjRotate(&object[obj_control], 0, 0, -20);
			break;

		case NSPageUpFunctionKey:
			obj_control++;
			if (obj_control >= obj_qty) obj_control = 0; 
			break;

		case NSPageDownFunctionKey:
			obj_control--;
			if (obj_control < 0) obj_control = obj_qty - 1;
			break;

		case 'q':
		case 'Q':
			if (FullScreen) 
				[NSApp terminate: self];
			break;
			
		default:
			break;
	}
}

- (void) render {
	
	int i, j;
	
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); // This clear the background color to dark blue
    glMatrixMode(GL_MODELVIEW); // Modeling transformation
    glLoadIdentity(); // Initialize the model matrix as identity
	
	glLoadMatrixf(&camera.matrix[0][0]); // At first we load the camera matrix
	
	glLightfv (GL_LIGHT1, GL_POSITION, light_position); // We don't want to move the light with the camera so we set again the light position after the camera transformation
	
    for (i = 0; i < obj_qty; i++)
	{
		glPushMatrix(); // We save the current matrix
		
		glMultMatrixf(&object[i].matrix[0][0]); // Now let's multiply the object matrix by the camera matrix
		
		if (object[i].id_texture != -1) 
		{
			glBindTexture(GL_TEXTURE_2D, object[i].id_texture); // We set the active texture 
		    glEnable(GL_TEXTURE_2D); // Texture mapping ON
		}
		else
		    glDisable(GL_TEXTURE_2D); // Texture mapping OFF
		
		glBegin(GL_TRIANGLES); // glBegin and glEnd delimit the vertices that define a primitive (in our case triangles)
		for (j = 0; j < object[i].polygons_qty; j++)
		{
			//----------------- FIRST VERTEX -----------------
			//Normal coordinates of the first vertex
			glNormal3f( object[i].normal[ object[i].polygon[j].a ].x,
					   object[i].normal[ object[i].polygon[j].a ].y,
					   object[i].normal[ object[i].polygon[j].a ].z);
			// Texture coordinates of the first vertex
			glTexCoord2f( object[i].mapcoord[ object[i].polygon[j].a ].u,
						 object[i].mapcoord[ object[i].polygon[j].a ].v);
			// Coordinates of the first vertex
			glVertex3f( object[i].vertex[ object[i].polygon[j].a ].x,
					   object[i].vertex[ object[i].polygon[j].a ].y,
					   object[i].vertex[ object[i].polygon[j].a ].z);
			
			//----------------- SECOND VERTEX -----------------
			//Normal coordinates of the second vertex
			glNormal3f( object[i].normal[ object[i].polygon[j].b ].x,
					   object[i].normal[ object[i].polygon[j].b ].y,
					   object[i].normal[ object[i].polygon[j].b ].z);
			// Texture coordinates of the second vertex
			glTexCoord2f( object[i].mapcoord[ object[i].polygon[j].b ].u,
						 object[i].mapcoord[ object[i].polygon[j].b ].v);
			// Coordinates of the second vertex
			glVertex3f( object[i].vertex[ object[i].polygon[j].b ].x,
					   object[i].vertex[ object[i].polygon[j].b ].y,
					   object[i].vertex[ object[i].polygon[j].b ].z);
			
			//----------------- THIRD VERTEX -----------------
			//Normal coordinates of the third vertex
			glNormal3f( object[i].normal[ object[i].polygon[j].c ].x,
					   object[i].normal[ object[i].polygon[j].c ].y,
					   object[i].normal[ object[i].polygon[j].c ].z);
			// Texture coordinates of the third vertex
			glTexCoord2f( object[i].mapcoord[ object[i].polygon[j].c ].u,
						 object[i].mapcoord[ object[i].polygon[j].c ].v);
			// Coordinates of the Third vertex
			glVertex3f( object[i].vertex[ object[i].polygon[j].c ].x,
					   object[i].vertex[ object[i].polygon[j].c ].y,
					   object[i].vertex[ object[i].polygon[j].c ].z);
		}
		glEnd();
		glPopMatrix(); // Restore the previous matrix 
	}
	
	[[NSOpenGLContext currentContext] flushBuffer];
}

- (void) initGL {
	
	// Call every time when current context was changed

	glClearColor(0.0, 0.0, 0.0, 0.0); // Clear background color to black
	glShadeModel(GL_SMOOTH); // Type of shading for the polygons 
	
	// Viewport transformation
    glViewport(0, 0, bounds.size.width, bounds.size.height);  
	
    // Projection transformation
    glMatrixMode(GL_PROJECTION); // Specifies which matrix stack is the target for matrix operations 
    glLoadIdentity(); // We initialize the projection matrix as identity
    gluPerspective(45.0f,(GLfloat)bounds.size.width/(GLfloat)bounds.size.height, 0.1f, 10000.0f); // We define the "viewing volume"
	glMatrixMode(GL_MODELVIEW);
	
	//Lights initialization and activation
    glLightfv (GL_LIGHT1, GL_AMBIENT, light_ambient);
    glLightfv (GL_LIGHT1, GL_DIFFUSE, light_diffuse);
    glLightfv (GL_LIGHT1, GL_DIFFUSE, light_specular);
    glLightfv (GL_LIGHT1, GL_POSITION, light_position);    
    glEnable (GL_LIGHT1);
    glEnable (GL_LIGHTING);
	
    //Materials initialization and activation
	glMaterialfv (GL_FRONT, GL_AMBIENT, mat_ambient);
    glMaterialfv (GL_FRONT, GL_DIFFUSE, mat_diffuse);
    glMaterialfv (GL_FRONT, GL_DIFFUSE, mat_specular);
    glMaterialfv (GL_FRONT, GL_POSITION, mat_shininess);
	
	// We enable the depth test (also called z buffer)
    glEnable(GL_DEPTH_TEST); 
	glClearDepth(1.0);
	glDepthFunc(GL_LESS);
	
	//Other initializations
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_CULL_FACE); // Enable the back face culling
	
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);
	
	if (filling == 1) 
		glPolygonMode (GL_FRONT_AND_BACK, GL_FILL); // Polygon rasterization mode (polygon filled)
    else 
		glPolygonMode (GL_FRONT_AND_BACK, GL_LINE); // Polygon rasterization mode (polygon outlined)
}

- (void) initScene {
	
	// Call at once to initialize global variables, function pointers, checking OpenGL 
	
	MatrGenerateLookupTab();

	MatrIdentity_4x4(camera.matrix); //Camera initialization
}

- (void) shareResources {
	
	// Call at once to initialize shared resources such as display lists, textures, buffer objects 
	// method is invoked after -initScene 
	
	NSString *object_name[] = {
		@"fighter1",
		@"fighter2",
		@"fighter3",
		nil
	};
	
	NSString *texture_name[] = { 
		@"skull",
		@"spaceshiptexture",
		@" "								// empty name
	}; 
	
	for (int x = -10; object_name[obj_qty]; obj_qty++, x += 10) { 
		
		NSString *object_name_path = [[NSBundle mainBundle] pathForResource: object_name[obj_qty] ofType: @"3ds"];
		
		if (!object_name_path) { 
			if (FullScreen) [controller exitFullScreen];
			Sys_CriticalAlert([NSString stringWithFormat: @"File \"%@.3ds\" does not exist", object_name[obj_qty]]);
			[NSApp terminate: self];
		}
		
		NSString *texture_name_path = [[NSBundle mainBundle] pathForResource: texture_name[obj_qty] ofType: @"bmp"];
		
		if (!texture_name_path && texture_name[obj_qty] != @" ") { 
			if (FullScreen) [controller exitFullScreen];
			Sys_CriticalAlert([NSString stringWithFormat: @"File \"%@\" does not exist", texture_name[obj_qty]]);
			[NSApp terminate: self];
		}
		
		int error = ObjLoad(&object[obj_qty], object_name_path, texture_name_path, (float) x, 0.0, -30.0, 900, 0, 0); 
		
		if (error == AF_ERROR_3DS) {
			if (FullScreen) [controller exitFullScreen];
			Sys_CriticalAlert([NSString stringWithFormat: @"\"%@.3ds\" is incorrect for more information see system log", object_name[obj_qty]]);
			[NSApp terminate: self];
		}
		
		if (error == AF_ERROR_TEXTURE) {
			if (FullScreen) [controller exitFullScreen];
			Sys_CriticalAlert([NSString stringWithFormat: @"\"%@\" is not a legal RGB / RGBA image, for more information see system log", texture_name[obj_qty]]);
			[NSApp terminate: self];
		}
	}
}

- (void) reshape {
	
	// Call every time when bounds where changed 
	
	glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); // We clear both the color and the depth buffer so to draw the next frame
    glViewport(0, 0, bounds.size.width, bounds.size.height); // Viewport transformation
	
    glMatrixMode(GL_PROJECTION); // Projection transformation
    glLoadIdentity(); // We initialize the projection matrix as identity
    gluPerspective(45.0f, (GLfloat)bounds.size.width/(GLfloat)bounds.size.height, 0.1f, 10000.0f); 
	glMatrixMode(GL_MODELVIEW);
}

- (void) setBounds: (NSRect) aFrame {
	
	bounds = aFrame;
	[self reshape];
}

@end
