//
//  AFController.m
//
//  Created by Ariel Feinerman on 1/28/10.
//  Copyright 2010 Alice Software Inc. All rights reserved.
//

#import "AFApplication.h"
#import "AFController.h"
#import "AFOpenGLView.h"
#import "AFTimer.h"
#import "Scene.h"

BOOL FullScreen = NO; // Global variable 

void Sys_CriticalAlert(NSString *text)
{
	NSAlert *alert = [NSAlert new];
	[alert setAlertStyle: NSCriticalAlertStyle];
	[alert setMessageText: @"Critical Error"];
	[alert setInformativeText: text];
	[alert addButtonWithTitle: @"Quit"];
	[alert runModal];
	[alert release];
}

NSOpenGLContext *Sys_CreateFullScreenContext(NSOpenGLContext *shareContext)
{
	
	NSOpenGLPixelFormatAttribute attrs [] = {
		NSOpenGLPFANoRecovery,
		NSOpenGLPFAFullScreen,
		NSOpenGLPFAScreenMask, CGDisplayIDToOpenGLDisplayMask(kCGDirectMainDisplay),
		NSOpenGLPFAColorSize, 32,
		NSOpenGLPFADepthSize, 32,
		NSOpenGLPFAMaximumPolicy,
		NSOpenGLPFADoubleBuffer,
		NSOpenGLPFAAccelerated,
		0
	};
	
	NSOpenGLPixelFormat *pixelFormat = [[NSOpenGLPixelFormat alloc] initWithAttributes: attrs];
	
	NSOpenGLContext *fullScreenContext = [[NSOpenGLContext alloc] initWithFormat: pixelFormat 
																	shareContext: shareContext];
	[pixelFormat release];
	pixelFormat = nil;
	
	if (fullScreenContext == nil) {
		NSLog(@"Failed to create fullScreen context");
		return nil;
	}
	
	return fullScreenContext;
}

NSOpenGLContext *Sys_CreateViewContext(NSOpenGLContext *shareContext) 
{
	NSOpenGLPixelFormatAttribute attrs [] = {
		NSOpenGLPFANoRecovery,
		NSOpenGLPFAWindow,
		NSOpenGLPFAColorSize, 32,
		NSOpenGLPFADepthSize, 32,
		NSOpenGLPFAMaximumPolicy,
		NSOpenGLPFADoubleBuffer,
		NSOpenGLPFAAccelerated,
		0
	};
	
	NSOpenGLPixelFormat *pixelFormat = [[NSOpenGLPixelFormat alloc] initWithAttributes: attrs];
	
	NSOpenGLContext *viewContext = [[NSOpenGLContext alloc] initWithFormat: pixelFormat 
																	shareContext: shareContext];
	[pixelFormat release];
	pixelFormat = nil;
	
	if (viewContext == nil) {
		NSLog(@"Failed to create view context");
		return nil;
	}
	
	return viewContext;
}

@interface AFController (Internal)

// Initialize GL Contexts
- (BOOL) initFullScreen;
- (BOOL) initView;

@end

@implementation AFController

- (void) awakeFromNib {
	
	view = [[AFOpenGLView alloc] initWithFrame: [win frame] pixelFormat: [AFOpenGLView defaultPixelFormat]];
	[win setContentView: view];
	
	scene = [[Scene alloc] initWithController: self];
	[view setScene: scene];
	
	[NSApp setEngine: scene];
	
	if (FullScreen) {
		
		[view setHidden: YES];
		if (![self initFullScreen]) {
			Sys_CriticalAlert(@"Cannot initialize full screen, for more information see system log");
			[NSApp terminate: self];
		}
		CGDisplayHideCursor(kCGDirectMainDisplay);
	}
	else {
		
		if (![self initView]) {
			Sys_CriticalAlert(@"Cannot initialize view, for more information see system log");
			[NSApp terminate: self];
		}
	}
	
	[scene initScene];
	[scene initGL];
	[scene shareResources];
	
	timer = [[AFTimer alloc] initWithTarget: FullScreen ? scene : view];
}

// Internal Methods

- (BOOL) initFullScreen {
		
	if (!fullScreenContext && !(fullScreenContext = Sys_CreateFullScreenContext(viewContext))) {
		return NO;
	}
	
	CGDisplayErr err = CGDisplayCapture(kCGDirectMainDisplay);
	
	if (err != CGDisplayNoErr) {
		NSLog(@"Failed to capture display");
		return NO;
	}
	
	[fullScreenContext setFullScreen];
	[fullScreenContext makeCurrentContext];
	
	[scene setBounds: [[NSScreen mainScreen] frame]];
	
	return YES;
}

- (BOOL) initView {
	
	if (!viewContext) {
		if ((viewContext = Sys_CreateViewContext(fullScreenContext))) {
			[view setOpenGLContext: viewContext];
			[viewContext setView: view];
		}
		else 
			return NO;
	}
	
	[viewContext makeCurrentContext];
	
	[scene setBounds: [view bounds]];
	
	return YES;
}

- (void) enterFullScreen {
	
	[view setHidden: YES];
	
	if (![self initFullScreen]) {
		Sys_CriticalAlert(@"Cannot initialize full screen, for more information see system log");
		[NSApp terminate: self];
	}

	CGDisplayHideCursor(kCGDirectMainDisplay);
	
	[scene initGL];
	
	[timer setTarget: scene];
	FullScreen = YES;
}

- (void) exitFullScreen {
	
	CGDisplayErr err = CGDisplayRelease(kCGDirectMainDisplay);
	
	[fullScreenContext clearDrawable];
	CGDisplayShowCursor(kCGDirectMainDisplay);
	
	if (err != CGDisplayNoErr) {
		Sys_CriticalAlert(@"Display was released with error");
		[NSApp terminate: self];
	}

	[view setHidden: NO];
	[win makeFirstResponder: view];
	
	if (![self initView]) {
		Sys_CriticalAlert(@"Cannot initialize view, for more information see system log");
		[NSApp terminate: self];
	}
	
	[scene initGL];
	
	[timer setTarget: view];
	FullScreen = NO;
}

- (void) dealloc {
	
	[scene release];
	[view release];
	[timer release];
	
	[fullScreenContext release];
	[viewContext release];
	
	[super dealloc];
}

@end
